Module: C

Defined in:
lib/castaddon.rb,
lib/castaddon/type.rb

Overview

Include the CAST gem, which provides a C99-parser. The gem is based on an intermediate representation in the form of an abstract syntax tree (AST). The gem provides C to AST and AST to C.

Defined Under Namespace

Classes: FloatLiteral, Index, IntLiteral, Node, NodeList, Type